@@ -7,7 +7,7 @@ use crate::error::SyntaxError;
 use either::Either;
 use log::trace;
 use num_bigint::BigInt as BigIntValue;
-use std::{fmt::Write, iter::FusedIterator};
+use std::fmt::Write;
 
 impl<'a, I: Input> Lexer<'a, I> {
     /// Reads an integer, octal integer, or floating-point number
@@ -32,7 +32,7 @@ impl<'a, I: Input> Lexer<'a, I> {
             0f64
         } else {
             // Use read_number_no_dot to support long numbers.
-            let (val, s) = self.read_number_no_dot_as_str(10)?;
+            let (val, s, not_octal) = self.read_number_no_dot_as_str(10)?;
             if self.input.cur() == Some('n') {
                 self.input.bump();
                 return Ok(Either::Right(s));
@@ -57,21 +57,20 @@ impl<'a, I: Input> Lexer<'a, I> {
                     // strict mode hates non-zero decimals starting with zero.
                     // e.g. 08.1 is strict mode violation but 0.1 is valid float.
 
-                    if val.fract() < 1e-10 {
-                        let d = digits(val.round() as u64, 10);
+                    if val.fract() == 0.0 {
+                        let val_str = s.to_string();
 
                         // if it contains '8' or '9', it's decimal.
-                        if d.clone().any(|v| v == 8 || v == 9) {
+                        if not_octal {
                             // Continue parsing
                             self.emit_strict_mode_error(start, SyntaxError::LegacyDecimal);
                         } else {
                             // It's Legacy octal, and we should reinterpret value.
-                            let val = u64::from_str_radix(&val.to_string(), 8)
-                                .expect("Does this can really happen?");
-                            let val = val
-                                .to_string()
-                                .parse()
-                                .expect("failed to parse numeric value as f64");
+                            let val =
+                                lexical::parse_radix::<f64, _>(&val_str, 8).unwrap_or_else(|err| {
+                                    panic!("failed to parse {} using `lexical`: {:?}", val_str, err)
+                                });
+
                             return self.make_legacy_octal(start, val).map(Either::Left);
                         }
                     }
@@ -156,7 +155,7 @@ impl<'a, I: Input> Lexer<'a, I> {
         self.bump(); // 0
         self.bump(); // x
 
-        let (val, s) = self.read_number_no_dot_as_str(radix)?;
+        let (val, s, _) = self.read_number_no_dot_as_str(radix)?;
         if self.eat(b'n') {
             return Ok(Either::Right(s));
         }
@@ -196,13 +195,17 @@ impl<'a, I: Input> Lexer<'a, I> {
 
     /// This can read long integers like
     /// "13612536612375123612312312312312312312312".
-    fn read_number_no_dot_as_str(&mut self, radix: u8) -> LexResult<(f64, BigIntValue)> {
+    ///
+    ///
+    /// Returned bool is `true` is there was `8` or `9`.
+    fn read_number_no_dot_as_str(&mut self, radix: u8) -> LexResult<(f64, BigIntValue, bool)> {
         debug_assert!(
             radix == 2 || radix == 8 || radix == 10 || radix == 16,
             "radix for read_number_no_dot should be one of 2, 8, 10, 16, but got {}",
             radix
         );
         let start = self.cur_pos();
+        let mut non_octal = false;
 
         let mut read_any = false;
 
@@ -212,6 +215,11 @@ impl<'a, I: Input> Lexer<'a, I> {
             radix,
             |total, radix, v| {
                 read_any = true;
+
+                if v == 8 || v == 9 {
+                    non_octal = true;
+                }
+
                 (f64::mul_add(total, radix as f64, v as f64), true)
             },
             &mut raw,
@@ -226,6 +234,7 @@ impl<'a, I: Input> Lexer<'a, I> {
             val,
             BigIntValue::parse_bytes(&raw.0.take().unwrap().as_bytes(), radix as _)
                 .expect("failed to parse string as a bigint"),
+            non_octal,
         ))
     }
 
@@ -245,18 +254,18 @@ impl<'a, I: Input> Lexer<'a, I> {
     /// When `len` is not zero, this
     /// will return `None` unless the integer has exactly `len` digits.
     pub(super) fn read_int(&mut self, radix: u8, len: u8, raw: &mut Raw) -> LexResult<Option<f64>> {
-        let mut count = 0;
+        let mut count = 0u16;
         let v = self.read_digits(
             radix,
             |opt: Option<f64>, radix, val| {
                 count += 1;
                 let total = opt.unwrap_or_default() * radix as f64 + val as f64;
-                (Some(total), count != len)
+                (Some(total), count != len as u16)
             },
             raw,
             true,
         )?;
-        if len != 0 && count != len {
+        if len != 0 && count != len as u16 {
             Ok(None)
         } else {
             Ok(v)
@@ -390,46 +399,6 @@ impl<'a, I: Input> Lexer<'a, I> {
     }
 }
 
-fn digits(value: u64, radix: u64) -> impl Iterator<Item = u64> + Clone + 'static {
-    debug_assert!(radix > 0);
-
-    #[derive(Clone, Copy)]
-    struct Digits {
-        n: u64,
-        divisor: u64,
-    }
-
-    impl Digits {
-        fn new(n: u64, radix: u64) -> Self {
-            let mut divisor = 1;
-            while n >= divisor * radix {
-                divisor *= radix;
-            }
-
-            Digits { n, divisor }
-        }
-    }
-
-    impl Iterator for Digits {
-        type Item = u64;
-
-        fn next(&mut self) -> Option<u64> {
-            if self.divisor == 0 {
-                None
-            } else {
-                let v = Some(self.n / self.divisor);
-                self.n %= self.divisor;
-                self.divisor /= 10;
-                v
-            }
-        }
-    }
-
-    impl FusedIterator for Digits {}
-
-    Digits::new(value, radix)
-}
-
 #[cfg(test)]
 mod tests {
     use super::{input::StringInput, *};
